use std::{fmt::Display, path::PathBuf};

use crate::run::{BENCH_BASE_PATH, BENCH_DATA_PATH, BENCH_BIN_PATH, BENCH_LOG_PATH};
use crate::run::args::TestType::*;

const CHUNK_MAX_BITS: u64 = 18;
const ADDRESS_MIN_BITS: u64 = 3;


#[derive(Debug, Copy, Clone)]
pub struct BenchArgs {
    pub data_args: DataArgs,
    pub bloom_filter_args: BloomFilterArgs,
    pub scan_args: ScanArgs
}

impl Display for BenchArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "BenchArgs({}, {}, {})", self.data_args, self.bloom_filter_args, self.scan_args)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct DataArgs {
    pub seed: u64,
    pub scan_subnet_size: u64,
    pub hit_rate: f64,
    pub entries: u64,
}

impl Display for DataArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Data(seed: {}, subnet: {}, hit rate: {})", self.seed, self.scan_subnet_size, self.hit_rate)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct BloomFilterArgs {
    pub test_type: TestType,
    pub address_bits: u64,
    pub chunk_address_bits: u64,
    pub hash_count: u64,
}

impl Display for BloomFilterArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "BloomFilter(type: {}, address bits: {}, hashes: {})",self.test_type, self.address_bits, self.hash_count)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct ScanArgs {
    pub rate: u64
}

impl Display for ScanArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Scan(rate: {})", self.rate)
    }
}

impl BenchArgs {
    pub fn rel_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(self.data_args.rel_path());
        path.push(self.bloom_filter_args.rel_path());
        path.push(self.scan_args.rel_path());
        return path;
    }

    pub fn wd_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_DATA_PATH);
        path.push(self.rel_path());
        return path;
    }

    pub fn bin_path(&self) -> PathBuf {
        Self::bin_bin_path(self.data_args, self.bloom_filter_args)
    }

    pub fn bin_bin_path(data_args: DataArgs, bloom_args: BloomFilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_BIN_PATH);
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn bin_log_path(data_args: DataArgs, bloom_args: BloomFilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_LOG_PATH);
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn bin_wd_path(data_args: DataArgs, bloom_args: BloomFilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_DATA_PATH);
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn log_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_LOG_PATH);
        path.push(self.rel_path());
        return path;
    }
}

impl DataArgs {
    pub fn from(seed: u64, scan_subnet_size: u64, hit_rate: f64) -> Self {
        let entries = (((1u64 << scan_subnet_size) as f64) * hit_rate).round() as u64;
        Self {
            seed,
            scan_subnet_size,
            hit_rate,
            entries
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(self.seed.to_string());
        path.push(self.scan_subnet_size.to_string());
        path.push(self.hit_rate.to_string());
        return path;
    }
}

impl BloomFilterArgs {
    pub fn from(data_args: DataArgs, test_type: TestType) -> BloomFilterArgs {
        match test_type {
            Normal(false_hit_rate) | BpfStats(false_hit_rate) => {
                let hash_count = (-false_hit_rate.log2()).round();
                let size = (data_args.entries as f64) * (hash_count/ 2f64.ln());
                let address_bits = size.log2().round() as u64;
                let address_bits = ADDRESS_MIN_BITS.max(address_bits);
                let chunk_address_bits = CHUNK_MAX_BITS.min(address_bits);

                BloomFilterArgs {
                    test_type,
                    address_bits,
                    chunk_address_bits,
                    hash_count: hash_count as u64
                }
            }
            Baseline | EmptyFilter => BloomFilterArgs {
                test_type,
                address_bits: ADDRESS_MIN_BITS,
                chunk_address_bits: ADDRESS_MIN_BITS,
                hash_count: 0
            },
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        match self.test_type {
            BpfStats(_) => PathBuf::from(format!("{}-{}-bpf", self.address_bits, self.hash_count)),
            Normal(_) | EmptyFilter => PathBuf::from(format!("{}-{}", self.address_bits, self.hash_count)),
            Baseline => PathBuf::from("baseline"),
        }
    }
}

impl ScanArgs {
    pub fn new(rate: u64) -> Self {
        Self {
            rate
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        PathBuf::from(self.rate.to_string())
    }
}

#[derive(Debug, Copy, Clone)]
pub enum TestType {
    Baseline, EmptyFilter, BpfStats(f64), Normal(f64)
}

impl Display for TestType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}
