use std::{path::PathBuf, time::Instant, fs::File, io::{Read, self, Write}, os::unix::prelude::AsRawFd, process};

use aya::{Bpf, maps::Array as BpfArray};
use anyhow::{anyhow, Context};
use aya::programs::{Xdp, XdpFlags};
use aya_log::BpfLogger;
use clap::Parser;
use env_logger::Env;
use log::{info, warn};
use tokio::signal;
use responder_common::bloom_filter;

#[cfg(all(debug_assertions, not(feature = "default_artifact_build")))]
const DEFAULT_TARGET: &str = "target/bpfel-unknown-none/debug";

#[cfg(all(not(debug_assertions), not(feature = "default_artifact_build")))]
const DEFAULT_TARGET: &str = "target/bpfel-unknown-none/release";

#[cfg(feature = "default_artifact_build")]
const DEFAULT_TARGET: &str = "bin/ebpf";

#[derive(Debug, Parser)]
struct Opt {
    #[clap(short, long, default_value = "lo")]
    iface: String,
    #[clap(short, long, default_value = "syn")]
    scan_type: String,
    #[clap(short, long)]
    bfb: Option<String>,
    #[clap(default_value = DEFAULT_TARGET, long)]
    target: PathBuf,
    #[clap(short, long)]
    fd_info_out_path: Option<PathBuf>
}

#[tokio::main]
async fn main() -> Result<(), anyhow::Error> {
    let start = Instant::now();
    let opt = Opt::parse();

    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    let mut bpf_path = PathBuf::new();
    bpf_path.push(opt.target);

    let xdp_name = opt.scan_type;
    bpf_path.push(xdp_name.clone());

    let mut bpf = Bpf::load_file(bpf_path)?;

    if let Err(e) = BpfLogger::init(&mut bpf) {
        warn!("failed to initialize eBPF logger: {}
This can happen if the loaded eBPF program has no log statements.", e);
    }

    // Obtain and load the XDP program called "responder" defined in the XDP file loaded above
    let program: &mut Xdp = bpf.program_mut("responder").unwrap().try_into()?;
    program.load()?;
    program.attach(&opt.iface, XdpFlags::default())
           .context(format!(
               "failed to attach the {} XDP program with default flags \
 - try changing XdpFlags::default() to XdpFlags::SKB_MODE",
               xdp_name
           ))?;

    info!("Loaded {} XDP program; FD: {}", xdp_name, program.fd().unwrap().as_raw_fd());


    if let Some(path) = opt.fd_info_out_path {
        info!("Emitting pid and fd to {:?}", &path);
        let info = format!("{{\"fd\": {}, \"pid\": {}}}\n", program.fd().unwrap().as_raw_fd(), process::id());
        File::create(path)?.write_all(info.as_bytes())?;
    }

    if let Some(bfb_path) = opt.bfb {
        info!("Installing filter rules from {}", bfb_path);
        let mut filter_map: BpfArray<_, [u8; bloom_filter::CHUNK_BYTES]> =
            BpfArray::try_from(bpf.map_mut("FILTER_MAP")
                               .ok_or(anyhow!("Could not construct mutable FILTER_MAP"))?)?;
        let mut bloom_filter_binary = File::open(bfb_path)?;
        let mut dimensions_buf: [u8; 8] = [0u8; 8];

        bloom_filter_binary.read_exact(&mut dimensions_buf)?;
        let file_address_bits = u64::from_be_bytes(dimensions_buf);
        if file_address_bits != (bloom_filter::ADDRESS_BITS as u64) {
            return Err(anyhow!(
                "Filter dimension ADDRESS_BITS does not match binary. Please rebuild the filter or recompile the binaries\n\tFile:\t{}\n\tBinary:\t{}",
                file_address_bits, bloom_filter::ADDRESS_BITS
            ));
        }

        bloom_filter_binary.read_exact(&mut dimensions_buf)?;
        let file_address_bits_chunk = u64::from_be_bytes(dimensions_buf);
        if file_address_bits_chunk != (bloom_filter::ADDRESS_BITS_CHUNK as u64) {
            return Err(anyhow!(
                "Filter dimension ADDRESS_BITS_CHUNK does not match binary. Please rebuild the filter or recompile the binaries\n\tFile:\t{}\n\tBinary:\t{}",
                file_address_bits_chunk, bloom_filter::ADDRESS_BITS_CHUNK
            ));
        }

        let mut chunk = [0 as u8; bloom_filter::CHUNK_BYTES];

        let mut i = 0;
        loop {
            let read = bloom_filter_binary.read(&mut chunk)?;
            if read > 0 {
                filter_map.set(i, chunk, 0)?;
                i += 1;
            }

            if read < bloom_filter::CHUNK_BYTES {
                break;
            }
        }
        if (i as usize) != bloom_filter::MAP_SIZE {
            return Err(anyhow!(
                "Number of chunks in file ({}) do not match expected number ({})",
                i, bloom_filter::MAP_SIZE
            ))
        }
    };

    info!("Loading XDP program took {:?}", start.elapsed());
    println!("press any key to exit");
    let mut buf = [0u8];
    io::stdin().read_exact(&mut buf)?;
    // signal::ctrl_c().await?;
    info!("Exiting...");

    Ok(())
}
